/*
 * Copyright 2014-present MongoDB, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef PHONGO_VERSION_H
#define PHONGO_VERSION_H

/* Note: this file should remain in the repository's root directory so that
 * PECL's release-upload.php script can verify the version information when
 * publishing a release. */

/* clang-format off */
#define PHP_MONGODB_VERSION "1.21.0"
#define PHP_MONGODB_STABILITY "stable"
#define PHP_MONGODB_VERSION_DESC 1,21,0,0
/* clang-format on */

#endif /* PHONGO_VERSION_H */
